import json
import os
from jinja2 import Template


def get_templates_text():
    template_dir = os.path.realpath(os.path.dirname(__file__) + '/templates')
    template_files = {'alpaca_index': 'alpaca_index.html',
                      'alpaca_form': 'alpaca_form.html',
                      'alpaca_header': 'alpaca_header.html',
                      'alpaca_example_data': 'alpaca_example.json'}

    paths = {key: "%s/%s" % (template_dir, file) for key, file in template_files.items()}
    templates = {key: open(filename, 'r').read() for key, filename in paths.items()}
    return templates


def form_data_to_form(form_data, templates):
    alpaca_json_data = json.dumps(form_data, indent=2)
    template = Template(templates['alpaca_form'])
    return template.render(alpaca_json_data=alpaca_json_data)


def form_data_to_form_elements(form_data):
    templates = get_templates_text()

    if form_data is None:
        templates_text = templates['alpaca_example_data']
        form_data = json.loads(templates_text)

    header_text = templates['alpaca_header']
    form_div = form_data_to_form(form_data, templates)

    return header_text, form_div


def form_data_to_html_page(form_data):
    header_text, form_div = form_data_to_form_elements(form_data)
    index_text = get_templates_text()['alpaca_index']
    index = Template(index_text)
    return index.render(alpaca_header=header_text, form_div=form_div)



from wickedhot import one_hot_encode as ohe
from random import Random


def test_get_one_hot_encoder_dicts_from_data_stream_with_ties():
    data = [{'letter': 'a', 'animal': 'dog', 'color': 'red'},
            {'letter': 'a', 'animal': 'dog', 'color': 'red'},
            {'letter': 'a', 'animal': 'cat', 'color': 'blue'},
            {'letter': 'b', 'animal': 'cat', 'color': 'blue'},
            {'letter': 'b', 'animal': 'dog', 'color': 'blue'},
            {'letter': 'b', 'animal': 'cat', 'color': 'green'}]

    stream = (i for i in data)

    categorical_n_levels_dict = {'letter': 1, 'animal': 1, 'color': 2}

    encoders = ohe.get_one_hot_encoder_dicts_from_data_stream(stream, categorical_n_levels_dict)

    expected = {'letter': {'a': 0},
                'animal': {'cat': 0},
                'color': {'blue': 0, 'red': 1}}

    assert encoders == expected


def test_get_key_val_pair_to_index_lookup_simple():
    encoder_dict = {'letter': {'a': 0},
                    'animal': {'cat': 0},
                    'color': {'blue': 0, 'red': 1}}

    non_encoded_keys = []
    index = ohe.get_key_val_pair_to_index_lookup(encoder_dict, non_encoded_keys)

    expected = {('animal', 'cat'): 0,
                ('color', 'blue'): 1,
                ('color', 'red'): 2,
                ('letter', 'a'): 3}

    assert index == expected


def test_get_key_val_pair_to_index_lookup_simple_with_numeric():
    encoder_dict = {'letter': {'a': 0},
                    'animal': {'cat': 0},
                    'color': {'blue': 0, 'red': 1}}

    non_encoded_keys = ['blah']
    index = ohe.get_key_val_pair_to_index_lookup(encoder_dict, non_encoded_keys)

    expected = {'blah': 0,
                ('animal', 'cat'): 1,
                ('color', 'blue'): 2,
                ('color', 'red'): 3,
                ('letter', 'a'): 4}

    assert index == expected


def test_get_line_encoder_and_decoder():
    encoder_dict = {'letter': {'a': 0},
                    'animal': {'cat': 0},
                    'color': {'blue': 0, 'red': 1}}

    non_encoded_keys = []
    index = ohe.get_key_val_pair_to_index_lookup(encoder_dict, non_encoded_keys)
    e, d = ohe.get_line_encoder_and_decoder(index)

    line = {'letter': 'a', 'animal': 'cat', 'color': 'blue'}
    encoded_line = e(line)
    expected = [1.0, 1.0, 0.0, 1.0]

    assert encoded_line == expected

    decoded_line = d(encoded_line)
    assert decoded_line == line


def test_get_line_encoder_and_decoder_with_numerics():
    encoder_dict = {'letter': {'a': 0},
                    'animal': {'cat': 0},
                    'color': {'blue': 0, 'red': 1}}

    non_encoded_keys = ['foo', 'bar']
    index = ohe.get_key_val_pair_to_index_lookup(encoder_dict, non_encoded_keys)
    e, d = ohe.get_line_encoder_and_decoder(index)

    line = {'foo': 88.0, 'letter': 'a', 'animal': 'cat', 'color': 'blue', 'bar': 5.5}
    encoded_line = e(line)
    expected = [5.5, 88.0, 1.0, 1.0, 0.0, 1.0]

    assert encoded_line == expected

    decoded_line = d(encoded_line)
    assert decoded_line == line


def test_get_one_hot_encoder_dicts_from_data_stream():
    data = [{'letter': 'a', 'animal': 'dog', 'color': 'red'},
            {'letter': 'b', 'animal': 'dog', 'color': 'red'},
            {'letter': 'b', 'animal': 'cat', 'color': 'blue'},
            {'letter': 'c', 'animal': 'cat', 'color': 'blue'},
            {'letter': 'a', 'animal': 'dog', 'color': 'blue'},
            {'letter': 'a', 'animal': 'mouse', 'color': 'green'}]

    stream = (i for i in data)

    categorical_n_levels_dict = {'letter': 2, 'animal': 5, 'color': 2}

    encoders = ohe.get_one_hot_encoder_dicts_from_data_stream(stream, categorical_n_levels_dict)

    expected = {'letter': {'a': 0, 'b': 1},
                'animal': {'dog': 0, 'cat': 1, 'mouse': 2},
                'color': {'blue': 0, 'red': 1}}

    assert encoders == expected


def test_get_numeric_stats():
    rand_generator = Random()
    rand_generator.seed(42)

    stream = [{'a': 0.0, 'b': 10.0, 'c': 100.0, 'd': 'foo'},
              {'a': 1.0, 'b': -10.0, 'c': 600.0, 'd': 'foo'},
              {'a': 5.0, 'b': 90.0, 'c': -1000.0, 'd': 'foo'}]

    numeric_stats = ohe.get_numeric_stats(stream, ['a', 'b', 'c'], rand_generator)

    expected = {'a': {'mean': 2.0,
                      'min': 0.0,
                      'max': 5.0,
                      'median': 1.0},
                'b': {'mean': 30.0,
                      'min': -10.0,
                      'max': 90.0,
                      'median': 10.0},
                'c': {'mean': -100.0,
                      'min': -1000,
                      'max': 600.0,
                      'median': 100.0}}

    assert numeric_stats == expected



from wickedhot.form_generator import encoder_package_to_html_page


def test_form_generation():

    stats = {'weight': {'median': 160.5, 'min': 55.0, 'max': 564.2},
             'height': {'median': 6.0,  'min': 4.5, 'max': 7.5}}

    # can also try with stats = None
    # stats = None

    packaged = {'max_levels_default': 10000,
                'numeric_cols': ['weight', 'height'],
                'numeric_stats': stats,
                'categorical_n_levels_dict': {'animal': 2, 'color': 1},
                'one_hot_encoder_dicts': {'animal': {'cat': 0, 'mouse': 1}, 'color': {'blue': 0}}}

    index_html = encoder_package_to_html_page(packaged)

    assert isinstance(index_html, str)
    assert "<html" in index_html

    # ensure default URL is not public
    assert "http://httpbin.org/post" not in index_html

    index_html = encoder_package_to_html_page(packaged, post_url='PUBLIC')

    assert isinstance(index_html, str)
    assert "<html" in index_html
    # ensure default URL is using public debugging option
    assert "http://httpbin.org/post" in index_html

    write_page = True
    if write_page:
        filename = 'junk_index.html'
        fp = open(filename, 'w')
        fp.write(index_html)
        fp.close()

        print("wrote: %s" % filename)





from wickedhot.one_hot_encode import unknown_level_value
from wickedhot.html_form import form_data_to_form_elements, form_data_to_html_page


def encoder_package_to_schema(encoder_package):

    properties = {}
    stats = encoder_package['numeric_stats']
    for field in encoder_package['numeric_cols']:
        properties[field] = {
            "type": "number",
            "title": field.capitalize(),
            "required": True
        }

        if stats is not None:
            properties[field]['minimum'] = stats[field]['min']
            properties[field]['maximum'] = stats[field]['max']

    encoder_dicts = encoder_package['one_hot_encoder_dicts']

    for field, value_dicts in encoder_dicts.items():
        values = sorted(value_dicts.items(), key=lambda x: x[1])
        levels = [v[0] for v in values]
        levels = levels + [unknown_level_value]

        properties[field] = {
            "type": "string",
            "title": field.capitalize(),
            "required": True,
            "enum": levels
        }

    schema = {
        "title": "Input features",
        "description": "Enter features",
        "type": "object",
        "properties": properties
    }

    return schema


def encoder_package_to_options(encoder_package, post_url=None):
    """
    :param encoder_package: one hot encoder package
    :param post_url: url to send form data to on submission
        default is ''
        for testing purposes, you may use PUBLIC and it will use
        "http://httpbin.org/post" which prints the result
        this is not secure so don't do that with sensitive data
    :return:
    """

    if post_url is None:
        post_url = ''

    if post_url == 'PUBLIC':
        post_url = "http://httpbin.org/post"

    fields = {}
    for field in encoder_package['numeric_cols']:
        fields[field] = {
            "size": 20,
            # "helper": "Please enter %s" % field
        }

    encoder_dicts = encoder_package['one_hot_encoder_dicts']

    for field, value_dicts in encoder_dicts.items():
        values = sorted(value_dicts.items(), key=lambda x: x[1])
        levels = [v[0] for v in values]
        levels = levels + [unknown_level_value]

        fields[field] = {
            "type": "select",
            # "helper": "Select %s" % field,
            "optionLabels": levels,
            "sort": False
        }

    options = {
        "form": {
            "attributes": {
                "action": post_url,
                "method": "post"
            },
            "buttons": {
                "submit": {}
            }
        },
        "helper": "Hit submit to update the prediction",
        "fields": fields}

    return options


def encoder_package_to_form_data(encoder_package, post_url=None):
    """
    Generate the form
    :param encoder_package: encoder package dict
    :param post_url: url to send form data to on submission
        default is ''
        for testing purposes, you may use PUBLIC and it will use
        "http://httpbin.org/post" which prints the result
        this is not secure so don't do that with sensitive data
    :return: form data
    """

    schema = encoder_package_to_schema(encoder_package)
    options = encoder_package_to_options(encoder_package, post_url=post_url)

    stats = encoder_package['numeric_stats']

    if stats is None:
        data = {field: 0 for field in encoder_package['numeric_cols']}
    else:
        data = {field: "%0.2f" % stats[field]['median'] for field in encoder_package['numeric_cols']}

    form_data = {"schema": schema,
                 "options": options,
                 "view": "bootstrap-edit",
                 "data": data}

    return form_data


def encoder_package_to_form_elements(encoder_package, post_url=None):
    form_data = encoder_package_to_form_data(encoder_package, post_url=post_url)
    header_text, form_div = form_data_to_form_elements(form_data)
    return header_text, form_div


def encoder_package_to_html_page(encoder_package, post_url=None):
    form_data = encoder_package_to_form_data(encoder_package, post_url=post_url)
    return form_data_to_html_page(form_data)



from wickedhot.one_hot_encoder import OneHotEncoder
from tempfile import NamedTemporaryFile


def test_init_empty():
    OneHotEncoder([], [])


def test_init():
    encoder = OneHotEncoder({'foo': 3, 'bar': 7}, ['buzz_numeric'])
    assert encoder.numeric_cols == ['buzz_numeric']
    assert encoder.categorical_n_levels_dict == {'foo': 3, 'bar': 7}


def test_init_ignore_default():
    encoder = OneHotEncoder({'foo': 3, 'bar': 7}, ['buzz_numeric'], max_levels_default=999)
    assert encoder.numeric_cols == ['buzz_numeric']
    assert encoder.categorical_n_levels_dict == {'foo': 3, 'bar': 7}


def test_init_do_not_ignore_default():
    encoder = OneHotEncoder(['foo', 'bar'], ['buzz_numeric'], max_levels_default=999)
    assert encoder.numeric_cols == ['buzz_numeric']
    assert encoder.categorical_n_levels_dict == {'foo': 999, 'bar': 999}


def test_load_from_data():
    encoder = OneHotEncoder(['animal', 'color'], ['weight'], max_levels_default=100)
    data = [{'animal': 'cat', 'color': 'blue', 'weight': 1.0},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0},
            {'animal': 'cat', 'color': 'blue', 'weight': 2.0},
            {'animal': 'cat', 'color': 'blue', 'weight': 0.0},
            {'animal': 'cat', 'color': 'blue', 'weight': 99.9}]

    assert encoder.encoder is None
    assert encoder.decoder is None
    assert encoder.one_hot_encoder_dicts is None

    encoder.load_from_data_stream(data)

    assert encoder.encoder is not None
    assert encoder.decoder is not None
    assert encoder.one_hot_encoder_dicts is not None


def test_load_from_data_encodes_data():
    encoder = OneHotEncoder(['animal', 'color'], ['weight'], max_levels_default=100)
    data = [{'animal': 'cat', 'color': 'blue', 'weight': 1.0},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0},
            {'animal': 'cat', 'color': 'blue', 'weight': 2.0},
            {'animal': 'cat', 'color': 'blue', 'weight': 0.0},
            {'animal': 'cat', 'color': 'blue', 'weight': 99.9}]

    encoder.load_from_data_stream(data)

    encoded_data = [encoder.encode_row(row) for row in data]
    assert len(encoded_data) == len(data)
    assert len(encoded_data[0]) != len(data[0])


def test_load_from_data_encodes_data_correctly():
    encoder = OneHotEncoder(['animal', 'color'], ['weight'], max_levels_default=100)

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0},
            {'animal': 'mouse', 'color': 'purple', 'weight': 0.0},
            {'animal': 'mouse', 'color': 'black', 'weight': 99.9}]

    encoder.load_from_data_stream(data)

    encoded_data = [encoder.encode_row(row) for row in data]
    assert len(encoded_data) == len(data)
    assert len(encoded_data[0]) != len(data[0])

    first_row = encoded_data[0]

    expected = [6.0,  # weight is numeric and comes first
                1.0,  # animal is first categorical and cat is the most common, first row is cat
                0.0,  # animal, mouse is next most common, not a mouse
                0.0,  # animal, dog and fish tied for frequency but dog first alphabetically
                0.0,  # animal, fish, cat is not a fish
                1.0,  # color is next categorical alphabetically and blue is most common, first row blue
                0.0,  # black
                0.0,  # magenta
                0.0,  # purple
                0.0,  # red
                0.0]  # yellow
    assert first_row == expected

    second_row = encoded_data[1]

    expected = [3.0,  # weight is numeric and comes first
                1.0,  # animal is first categorical and cat is the most common, first row is cat
                0.0,  # animal, mouse is next most common, not a mouse
                0.0,  # animal, dog and fish tied for frequency but dog first alphabetically
                0.0,  # animal, fish, cat is not a fish
                0.0,  # color is next categorical alphabetically and blue is most common, first row blue
                0.0,  # black next alphabetically for ones with frequency 1
                0.0,  # magenta next
                0.0,  # purple
                1.0,  # red, this is red
                0.0]  # yellow
    assert second_row == expected

    last_row = encoded_data[-1]

    expected = [99.9,  # weight is numeric and comes first
                0.0,  # animal is first categorical and cat is the most common, first row is cat
                1.0,  # animal, mouse is next most common, not a mouse
                0.0,  # animal, dog and fish tied for frequency but dog first alphabetically
                0.0,  # animal, fish, cat is not a fish
                0.0,  # color is next categorical alphabetically and blue is most common, first row blue
                1.0,  # black next alphabetically for ones with frequency 1, this one black
                0.0,  # magenta next
                0.0,  # purple
                0.0,  # red
                0.0]  # yellow

    expected_total = [[6.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                      [3.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                      [5.5, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                      [7.0, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                      [2.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                      [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                      [99.9, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0]]

    assert encoded_data == expected_total


def test_save_load():
    filename = NamedTemporaryFile().name

    encoder = OneHotEncoder(['animal', 'color'], ['weight'], max_levels_default=100)

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0},
            {'animal': 'mouse', 'color': 'purple', 'weight': 0.0},
            {'animal': 'mouse', 'color': 'black', 'weight': 99.9}]

    encoder.load_from_data_stream(data)

    encoded_data = encoder.encode_data(data)

    encoder.save(filename)

    encoder_from_file = OneHotEncoder([], [])
    encoder_from_file.load_from_file(filename)

    encoded_data_from_file = encoder_from_file.encode_data(data)

    assert encoded_data == encoded_data_from_file


def test_inversion():
    encoder = OneHotEncoder(['animal', 'color'], ['weight'], max_levels_default=100)

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0},
            {'animal': 'mouse', 'color': 'purple', 'weight': 0.0},
            {'animal': 'mouse', 'color': 'black', 'weight': 99.9}]

    encoder.load_from_data_stream(data)

    encoded_data = encoder.encode_data(data)

    data_decoded = encoder.decode_data(encoded_data)
    assert data_decoded == data

    data_recoded = encoder.encode_data(data_decoded)
    assert data_recoded == encoded_data


def test_inversion_more_complicated():
    encoder = OneHotEncoder(['animal', 'color'], ['weight', 'height'], max_levels_default=100)

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0, 'height': 88.9},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0, 'height': 44.9},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5, 'height': 2.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0, 'height': 3233.2},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0, 'height': 666.6},
            {'animal': 'mouse', 'color': 'red', 'weight': 0.0, 'height': 55.5},
            {'animal': 'mouse', 'color': 'blah', 'weight': 99.9, 'height': 33}]

    encoder.load_from_data_stream(data)

    encoded_data = encoder.encode_data(data)

    data_decoded = encoder.decode_data(encoded_data)
    assert data_decoded == data

    data_recoded = encoder.encode_data(data_decoded)
    assert data_recoded == encoded_data


def test_inversion_more_complicated_with_max_levels():
    encoder = OneHotEncoder({'animal': 2, 'color': 2}, ['weight', 'height'])

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0, 'height': 88.9},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0, 'height': 44.9},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5, 'height': 2.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0, 'height': 3233.2},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0, 'height': 666.6},
            {'animal': 'mouse', 'color': 'red', 'weight': 0.0, 'height': 55.5},
            {'animal': 'mouse', 'color': 'blah', 'weight': 99.9, 'height': 33}]

    encoder.load_from_data_stream(data)

    encoded_data = encoder.encode_data(data)
    data_decoded = encoder.decode_data(encoded_data)

    expected = [{'height': 88.9, 'weight': 6.0, 'animal': 'cat', 'color': 'blue'},
                {'height': 44.9, 'weight': 3.0, 'animal': 'cat', 'color': 'red'},
                {'height': 2.5, 'weight': 5.5, 'color': 'UNKNOWN_CATEGORICAL_LEVEL', 'animal': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 3233.2, 'weight': 7.0, 'color': 'blue', 'animal': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 666.6, 'weight': 2.0, 'animal': 'cat', 'color': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 55.5, 'weight': 0.0, 'animal': 'mouse', 'color': 'red'},
                {'height': 33, 'weight': 99.9, 'animal': 'mouse', 'color': 'UNKNOWN_CATEGORICAL_LEVEL'}]

    assert data_decoded == expected


def run_example(stats=False):
    encoder = OneHotEncoder({'animal': 2, 'color': 1}, ['weight', 'height'])

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0, 'height': 88.9, 'extra_junk': 'blah'},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0, 'height': 44.9},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5, 'height': 2.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0, 'height': 3233.2},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0, 'height': 666.6},
            {'animal': 'mouse', 'color': 'red', 'weight': 0.0, 'height': 55.5},
            {'animal': 'mouse', 'color': 'blah', 'weight': 99.9, 'height': 33}]

    encoder.load_from_data_stream(data)

    encoded_data = encoder.encode_data(data)
    data_decoded = encoder.decode_data(encoded_data)

    expected = [{'height': 88.9, 'weight': 6.0, 'animal': 'cat', 'color': 'blue'},
                {'height': 44.9, 'weight': 3.0, 'animal': 'cat', 'color': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 2.5, 'weight': 5.5, 'color': 'UNKNOWN_CATEGORICAL_LEVEL', 'animal': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 3233.2, 'weight': 7.0, 'color': 'blue', 'animal': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 666.6, 'weight': 2.0, 'animal': 'cat', 'color': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 55.5, 'weight': 0.0, 'animal': 'mouse', 'color': 'UNKNOWN_CATEGORICAL_LEVEL'},
                {'height': 33, 'weight': 99.9, 'animal': 'mouse', 'color': 'UNKNOWN_CATEGORICAL_LEVEL'}]

    assert data_decoded == expected

    # add number stats?
    if stats:
        encoder.add_numeric_stats(data)

    # check the package
    packaged = encoder.package_data()
    return packaged


def test_inversion_more_complicated_with_max_levels_diff():
    packaged = run_example(stats=False)

    expected = {'max_levels_default': 10000,
                'numeric_cols': ['weight', 'height'],
                'categorical_n_levels_dict': {'animal': 2, 'color': 1},
                'one_hot_encoder_dicts': {'animal': {'cat': 0, 'mouse': 1}, 'color': {'blue': 0}},
                'numeric_stats': None}

    assert packaged == expected


def test_inversion_more_complicated_with_max_levels_diff_and_numeric_stats():
    packaged = run_example(stats=True)

    stats = {'weight': {'min': 0.0, 'max': 99.9, 'mean': 17.62857142857143, 'median': 5.5},
             'height': {'min': 2.5, 'max': 3233.2, 'mean': 589.2285714285715, 'median': 55.5}}

    expected = {'max_levels_default': 10000,
                'numeric_cols': ['weight', 'height'],
                'categorical_n_levels_dict': {'animal': 2, 'color': 1},
                'one_hot_encoder_dicts': {'animal': {'cat': 0, 'mouse': 1}, 'color': {'blue': 0}},
                'numeric_stats': stats}

    assert packaged == expected


def test_html_form():
    encoder = OneHotEncoder({'animal': 2, 'color': 1}, ['weight', 'height'])

    data = [{'animal': 'cat', 'color': 'blue', 'weight': 6.0, 'height': 88.9, 'extra_junk': 'blah'},
            {'animal': 'cat', 'color': 'red', 'weight': 3.0, 'height': 44.9},
            {'animal': 'dog', 'color': 'yellow', 'weight': 5.5, 'height': 2.5},
            {'animal': 'fish', 'color': 'blue', 'weight': 7.0, 'height': 3233.2},
            {'animal': 'cat', 'color': 'magenta', 'weight': 2.0, 'height': 666.6},
            {'animal': 'mouse', 'color': 'red', 'weight': 0.0, 'height': 55.5},
            {'animal': 'mouse', 'color': 'blah', 'weight': 99.9, 'height': 33}]

    encoder.load_from_data_stream(data)
    package = encoder.package_data()

    expected = {'max_levels_default': 10000,
                'numeric_cols': ['weight', 'height'],
                'categorical_n_levels_dict': {'animal': 2, 'color': 1},
                'one_hot_encoder_dicts': {'animal': {'cat': 0, 'mouse': 1}, 'color': {'blue': 0}},
                'numeric_stats': None}

    assert package == expected

    html_header, form_div = encoder.get_form_html_elements()

    form_tags = ['id="form"', 'schema', 'alpaca', 'script']

    for tag in form_tags:
        assert tag in form_div

    header_tags = ['jquery', 'bootstrap', 'alpaca', 'script']

    for tag in header_tags:
        assert tag in html_header

    html_page = encoder.get_form_html_page()

    assert "<html" in html_page
    assert html_header in html_page
    assert form_div in html_page



import json
from random import Random
from wickedhot import one_hot_encode as ohe
from wickedhot.form_generator import encoder_package_to_form_elements, encoder_package_to_html_page


class OneHotEncoder:
    def __init__(self, categorical_cols, numeric_cols, max_levels_default=10000):
        assert max_levels_default > 0
        self.max_levels_default = max_levels_default
        self.numeric_cols = numeric_cols
        self.one_hot_encoder_dicts = None
        self.encoder = None
        self.decoder = None
        self.index_lookup = None
        self.numeric_stats = None
        if isinstance(categorical_cols, list):
            self.categorical_n_levels_dict = {k: self.max_levels_default for k in categorical_cols}
        elif isinstance(categorical_cols, dict):
            assert min(list(categorical_cols.values())) > 0
            self.categorical_n_levels_dict = categorical_cols
        else:
            raise ValueError('categorical_cols must be a list or dictionary')

    def load_from_data_stream(self, stream_of_dicts):
        self.one_hot_encoder_dicts = ohe.get_one_hot_encoder_dicts_from_data_stream(stream_of_dicts,
                                                                                    self.categorical_n_levels_dict)
        self._get_encoder_decoder()

    def add_numeric_stats(self, stream_of_dicts):

        random_generator = Random()
        random_generator.seed(265472)

        self.numeric_stats = ohe.get_numeric_stats(stream_of_dicts, self.numeric_cols, random_generator)

    def package_data(self):
        data = {'max_levels_default': self.max_levels_default,
                'numeric_cols': self.numeric_cols,
                'categorical_n_levels_dict': self.categorical_n_levels_dict,
                'one_hot_encoder_dicts': self.one_hot_encoder_dicts,
                'numeric_stats': self.numeric_stats}

        return data

    def save(self, json_file_name):
        with open(json_file_name, 'w') as fp:
            json.dump(self.package_data(), fp)

    def load_from_packaged_data(self, data_object):
        self.max_levels_default = data_object['max_levels_default']
        self.numeric_cols = data_object['numeric_cols']
        self.one_hot_encoder_dicts = data_object['one_hot_encoder_dicts']

        self._get_encoder_decoder()

    def load_from_file(self, json_file_name):
        with open(json_file_name, 'r') as fp:
            packaged_data = json.load(fp)

        self.load_from_packaged_data(packaged_data)

    def _get_encoder_decoder(self):
        self.index_lookup = ohe.get_key_val_pair_to_index_lookup(self.one_hot_encoder_dicts, self.numeric_cols)
        self.index_lookup_rev = {v: k for k, v in self.index_lookup.items()}
        self.encoder, self.decoder = ohe.get_line_encoder_and_decoder(self.index_lookup)

    def encode_row(self, row):
        return self.encoder(row)

    def decode_row(self, row):
        return self.decoder(row)

    def index_to_column(self, index):
        return self.index_lookup_rev[index]

    def get_index(self, x):
        if isinstance(x, tuple):
            key, value = x
        elif isinstance(x, str):
            key = x
            value = None
        else:
            raise ValueError('x must be a string for numeric col of key value pair for categorical level')

        idx, _ = ohe.get_index(key, value, self.index_lookup)
        return idx

    def encode_data_stream(self, stream):
        # generator
        return (self.encode_row(row) for row in stream)

    def encode_data(self, stream):
        return list(self.encode_data_stream(stream))

    def decode_data_stream(self, encoded_data_stream):
        return (self.decode_row(row) for row in encoded_data_stream)

    def decode_data(self, encoded_data_stream):
        return list(self.decode_data_stream(encoded_data_stream))

    def get_form_html_elements(self, post_url=None):
        """
        Get html header text and form div text to be injected into an html page
        via templating (in the header section and body respectively)
        :param post_url: url that form will be submitted to on submission
        :return: html_header, div_text
        """
        return encoder_package_to_form_elements(self.package_data(), post_url=post_url)

    def get_form_html_page(self, post_url=None):
        """
        Return entire functioning html page using the default simple index.html template
        :param post_url: url that form will be submitted to on submission
        :return:
        """
        return encoder_package_to_html_page(self.package_data(), post_url=post_url)



from collections import Counter
from statistics import median
from wickedhot.res_sample import update_res_samples, initialize_res_samples

unknown_level_value = 'UNKNOWN_CATEGORICAL_LEVEL'


def one_hot_encode_column(data_stream, max_levels):
    count = Counter()
    for value in data_stream:
        count[value] += 1

    most_common = count.most_common()[0:max_levels-1]
    encoder_dict = {}
    i = 1
    for val, num in most_common:
        encoder_dict[val] = i
        i += 1


def select_counter(counter, n_max):
    # first by frequency (largest), then alphabetical by key
    return sorted(counter.most_common(), key=lambda x: (-x[1], x[0]))[0:n_max]


def get_one_hot_encoder_dicts_from_data_stream(stream_of_dicts, categorical_n_levels_dict):
    """
    :param stream_of_dicts: a stream of dictionaries (e.g from csv.DictReader)
    :param categorical_n_levels_dict: dictionary mapping categorical keys to
        maximum number of levels
    :return: dict of dicts. Maps [key][val] to an an index from 0 to max levels +1
            The number of 0 corresponds to either an uncommon level or unseen level
            which are treated the same.
    """
    # data stream is a stream of dictionaries as if from csv.DictReader
    keys = categorical_n_levels_dict.keys()
    count = {k: Counter() for k in keys}
    for row in stream_of_dicts:
        for k in keys:
            count[k][row[k]] += 1

    most_commons = {k: select_counter(count[k], n_max) for k, n_max in categorical_n_levels_dict.items()}
    return {k: {val[0]: i for i, val in enumerate(most_commons[k])} for k in keys}


def get_key_val_pair_to_index_lookup(encoder_dict, non_encoded_keys):
    encoded_keys = sorted(encoder_dict.keys())
    non_enc_keys = sorted(non_encoded_keys)

    index = {k: i for i, k in enumerate(non_enc_keys)}

    first = len(non_enc_keys)
    for key in encoded_keys:
        n = len(encoder_dict[key])
        for val, idx in encoder_dict[key].items():
            pair = (key, val)
            index[pair] = idx + first
        first += n

    return index


def get_index(key, val, index_lookup):
    idx = None
    encoded_value = None

    if key in index_lookup:
        # non-encoded
        idx = index_lookup[key]
        encoded_value = val

    if (key, val) in index_lookup:
        assert key not in index_lookup
        # one-hot-encoded
        idx = index_lookup[(key, val)]
        encoded_value = 1.0

    return idx, encoded_value


def get_line_encoder_and_decoder(index):
    n_index = len(index)
    all_cols = get_all_cols(index)
    rev_index = {v: k for k, v in index.items()}
    assert len(rev_index) == n_index

    def encoder(line):
        encoded_line = [0.0] * n_index
        for k, v in line.items():
            idx, encoded_value = get_index(k, v, index)
            if idx is not None:
                encoded_line[idx] = encoded_value

        return encoded_line

    def decoder(encoded_line):
        assert len(encoded_line) == n_index
        decoded_line = {}
        for idx, val in enumerate(encoded_line):
            value = rev_index[idx]

            if isinstance(value, tuple):
                assert val == 1.0 or val == 0.0
                if val == 1.0:
                    assert value[0] not in decoded_line
                    decoded_line[value[0]] = value[1]

            elif isinstance(value, str):
                decoded_line[value] = val

        fill_in_unknown_levels(decoded_line, all_cols)
        return decoded_line

    return encoder, decoder


def get_all_cols(index):
    all_cols = set()
    for k in index.keys():
        if isinstance(k, str):
            all_cols.add(k)
        if isinstance(k, tuple):
            all_cols.add(k[0])
    return all_cols


def fill_in_unknown_levels(decoded_line, all_cols):
    for k in all_cols:
        if k not in decoded_line:
            decoded_line[k] = unknown_level_value


def get_numeric_stats(stream_of_dicts, numeric_cols, random_generator):
    num_samples = 100
    res_samples = initialize_res_samples(numeric_cols)

    numeric_stats = {}
    for field in numeric_cols:
        numeric_stats[field] = {'sum': 0.0,
                                'min': 1e12,
                                'max': -1e12}

    num = 0
    for num, row in enumerate(stream_of_dicts):
        for field in numeric_cols:
            value = row[field]
            numeric_stats[field]['sum'] += value
            numeric_stats[field]['min'] = min(value, numeric_stats[field]['min'])
            numeric_stats[field]['max'] = max(value, numeric_stats[field]['max'])

            update_res_samples(res_samples, num, field, value, num_samples, random_generator)

    num += 1

    for field in numeric_cols:
        numeric_stats[field]['mean'] = numeric_stats[field]['sum']/num
        numeric_stats[field]['median'] = median(res_samples[field])
        del numeric_stats[field]['sum']

    return numeric_stats



"""
Reservoir sampling. Sample from stream of unknown length

usage:

from statistics import median

res = initialize_res_samples(fields)
num_samples = 50

random_generator = random.Random()
random_generator.seed(42)

for row_index, row in enumerate(stream):
    for field in fields:
        value = row[field]

    update_res_samples(res_samples, row_index, field, value, num_samples, random_generator)

medians = {field: median(res_samples[field] for field in fields}
"""


def initialize_res_samples(fields):
    return {field: [] for field in fields}


def update_res_samples(res_samples, row_index, field, value, num_samples, random_generator):
    """
    Update reservoir samples
    :param res_samples: dicts of lists, instantiated with initialize_res_samples
    :param row_index: the index in the stream
    :param field: field that is being updated
    :param value: value for this field in this row
    :param num_samples: number of samples to keep for each field
    :param random_generator: random.Random() instance, be sure to set seed
    :return:
    """
    if row_index + 1 <= num_samples:
        res_samples[field].append(value)
    else:
        probability = num_samples/(row_index + 1)
        ramdom_number = random_generator.random()
        if ramdom_number < probability:
            index = random_generator.randint(0, num_samples)
            res_samples[field][index] = value
